type bstmt =
    Lit of Types.t    (* Push a literal *)
  | Drp               (* Discard a value *)
  | Bin of Ast.op     (* Perform arithmetic on top of stack *)
  | Lod of int        (* Fetch global variable *)
  | Str of int        (* Store global variable *)
  | Lfp of int        (* Load frame pointer relative *)
  | Sfp of int        (* Store frame pointer relative *)
  | Jsr of int        (* Call function by absolute address *)
  | Ent of int        (* Push FP, FP -> SP, SP += i *)
  | Rts of int        (* Restore FP, SP, consume formals, push result *)
  | Beq of int        (* Branch relative if top-of-stack is zero *)
  | Bne of int        (* Branch relative if top-of-stack is non-zero *)
  | Bra of int        (* Branch relative *)
  | Lis of int        (* List construction *)
  | Has of int        (* Hash table construction *)
  | Rsn of int        (* Call restriction function by absolute address *)
  | Sel of Ast.select (* Selector chain literal *)
  | Slc of int        (* Selector construction *)
  | Nit of int        (* Next iteration *)
  | Hlt               (* Terminate *)

type prog = {
    num_globals : int;   (* Number of global variables *)
    text : bstmt array; (* Code for all the functions *)
  }

let string_of_stmt = function
    Lit(i) -> "Lit " ^ Types.string_of_typed_expr i
  | Drp -> "Drp"
  | Bin(Ast.Add) -> "Add"
  | Bin(Ast.Sub) -> "Sub"
  | Bin(Ast.Mult) -> "Mul"
  | Bin(Ast.Div) -> "Div"
  | Bin(Ast.Equal) -> "Eql"
  | Bin(Ast.Neq) -> "Neq"
  | Bin(Ast.Less) -> "Lt"
  | Bin(Ast.Leq) -> "Leq"
  | Bin(Ast.Greater) -> "Gt"
  | Bin(Ast.Geq) -> "Geq"
  | Lod(i) -> "Lod " ^ string_of_int i
  | Str(i) -> "Str " ^ string_of_int i
  | Lfp(i) -> "Lfp " ^ string_of_int i
  | Sfp(i) -> "Sfp " ^ string_of_int i
  | Jsr(i) -> "Jsr " ^ string_of_int i
  | Ent(i) -> "Ent " ^ string_of_int i
  | Rts(i) -> "Rts " ^ string_of_int i
  | Bne(i) -> "Bne " ^ string_of_int i
  | Beq(i) -> "Beq " ^ string_of_int i
  | Bra(i) -> "Bra " ^ string_of_int i
  | Lis(i) -> "Lis " ^ string_of_int i
  | Has(i) -> "Has " ^ string_of_int i
  | Rsn(i) -> "Rsn " ^ string_of_int i
  | Sel(Ast.Colon) -> "Co"
  | Sel(Ast.DoubleColon) -> "DCo"
  | Slc(i) -> "Slc " ^ string_of_int i
  | Nit(i) -> "Nit " ^ string_of_int i
  | Hlt    -> "Hlt"

let string_of_prog p =
  string_of_int p.num_globals ^ " global variables\n" ^
  let funca = Array.mapi
      (fun i s -> string_of_int i ^ " " ^ string_of_stmt s) p.text
  in String.concat "\n" (Array.to_list funca)
