open Bytecode

module NameMap = Map.Make(struct
    type t = string
    let compare x y = Pervasives.compare x y
end)

module StringMap = Map.Make(String)

(*
  implement the scope rule
  an assignment at toplevel creates a global variable if it is not defined yet.
  an assignment at a fun_decl, however, creates a local variable when the left
  hand side is neither a global nor formal let variable.
  a formal variable shadows a global variable of the same name
*)
let rec env_expr env formals toplevel = function
      Ast.Assign(var, expr) -> let globals, locals, funcs = env in
                               if List.mem var formals || List.mem var globals then
                                   env
                               else if toplevel then
                                   var :: globals, locals, funcs
                               else
                                   globals, var :: locals, funcs
    | _                     -> env

and env_stmt env formals toplevel = function
      Ast.Expr(e)       -> let env = env_expr env formals toplevel e in env
    | Ast.Block(stmts)  -> List.fold_left (fun env s -> env_stmt env formals toplevel s) env stmts
    | Ast.If(e,s1,s2)   -> let env = env_expr env formals toplevel e in 
                           let env = env_stmt env formals toplevel s1 in
                           let env = env_stmt env formals toplevel s2 in
                           env
    | Ast.While(e,s)    -> let env = env_expr env formals toplevel e in
                           let env = env_stmt env formals toplevel s in env
    | Ast.Foreach(e1,e2,s) -> let globals, locals, funcs = env in
                              let env = (match e1 with 
                                  Ast.Id id -> if toplevel then id::globals, locals, funcs
                                               else globals, id :: locals, funcs
                                | _ -> raise (Failure "foreach expects an identifier"))
                              in env_stmt env formals toplevel s
    | _                 -> env

and env_gstmt env = function
      Ast.Stmt(s)     -> let env = env_stmt env [] true s in env
    | Ast.Func(fdecl) -> let env = env_fdecl env fdecl.Ast.formals fdecl in env

and env_fdecl env formals fdecl = 
    let globals, localvars, funcs = List.fold_left (fun env s -> env_stmt env formals false s) env fdecl.Ast.body in 
    let funcs = NameMap.add fdecl.Ast.fname {fdecl with Ast.locals=localvars} funcs in
    globals, [], funcs                          

(* Symbol table: Information about all the names in scope *)
type env = {
    function_index : int StringMap.t; (* Index for each function *)
    global_index   : int StringMap.t; (* "Address" for global variables *)
    local_index    : int StringMap.t; (* FP offset for args, locals *)
  }

(* val enum : int -> 'a list -> (int * 'a) list *)
let rec enum stride n = function
    [] -> []
  | hd::tl -> (n, hd) :: enum stride (n+stride) tl

(* val string_map_pairs StringMap 'a -> (int * 'a) list -> StringMap 'a *)
let string_map_pairs map pairs =
  List.fold_left (fun m (i, n) -> StringMap.add n i m) map pairs

(** Translate a program in AST form into a bytecode program.  Throw an
    exception if something is wrong, e.g., a reference to an unknown
    variable or function *)
let translate gstmts = 
  let (globals, _, funcs) = List.fold_left env_gstmt ([], [], NameMap.empty) gstmts in
  let functions = NameMap.fold (fun k v a -> v :: a) funcs [] in

(*let translate (globals, functions) =*)

  (* Allocate "addresses" for each global variable *)
  let global_indexes = string_map_pairs StringMap.empty (enum 1 0 globals) in

  (* Assign indexes to function names; built-in "print" is special *)
  let built_in_functions = StringMap.add "print" (-1) StringMap.empty in
  let built_in_functions = StringMap.add "gets" (-2) built_in_functions in
  let built_in_functions = StringMap.add "parse" (-3) built_in_functions in
  let function_indexes = string_map_pairs built_in_functions
      (enum 1 1 (List.map (fun f -> f.Ast.fname) functions)) in

  let rec expr env = function
      Ast.Literal i -> [Lit (Types.Int i)]
    | Ast.Float f   -> [Lit (Types.Float f)]
    | Ast.Str s     -> [Lit (Types.Str s)]
    | Ast.Boolean b -> [Lit (Types.Boolean b)]
    | Ast.Undef     -> [Lit (Types.Undef)]
    | Ast.Lis l     -> List.concat (List.map (expr env) l) @ [Lis (List.length l)]
    | Ast.Hash h    -> List.concat (List.map (fun (k,v) ->(expr env v) @ [Lit (Types.Str k)]) h) @ [Has (List.length h)]
    | Ast.Id s ->
        (try [Lfp (StringMap.find s env.local_index)]
        with Not_found -> try [Lod (StringMap.find s env.global_index)]
        with Not_found -> raise (Failure ("undeclared variable " ^ s)))
    | Ast.Binop (e1, op, e2) -> expr env e1 @ expr env e2 @ [Bin op]
    | Ast.Assign (s, e) -> expr env e @
        (try [Sfp (StringMap.find s env.local_index)]
        with Not_found -> try [Str (StringMap.find s env.global_index)]
        with Not_found -> raise (Failure ("undeclared variable " ^ s)))
    | Ast.Select(id, chaining, postfix, restrictions) ->
      let rsn = (List.fold_left (fun acc fname -> [Rsn
				 (try StringMap.find fname env.function_index
				  with Not_found -> raise (Failure ("undefined function " ^ fname)))
			     ]@acc) [] (List.rev restrictions)) in rsn @ [Lis (List.length restrictions)] @
	[Lit (Types.Str postfix)] @
	List.concat (List.map (fun (k,v) ->(expr env v) @ [Sel k]) (List.rev chaining)) @
	(expr env id) @ [Slc (List.length chaining)]
    | Ast.Call (fname, actuals) -> (try
        (List.concat (List.map (expr env) (List.rev actuals))) @
        [Jsr (StringMap.find fname env.function_index) ]   
        with Not_found -> raise (Failure ("undefined function " ^ fname)))
    (*| Ast.Noexpr -> []*)

  in let rec stmt env num_formals = function
      Ast.Block sl     ->  List.concat (List.map (stmt env num_formals) sl)
    | Ast.Expr e       -> expr env e @ [Drp]
    | Ast.Return e     -> expr env e @ [Rts num_formals]
    | Ast.If (p, t, f) -> let t' = stmt env num_formals t and f' = stmt env num_formals f in
        expr env p @ [Beq(2 + List.length t')] @
        t' @ [Bra(1 + List.length f')] @ f'
    (*| Ast.For (e1, e2, e3, b) ->
        stmt (Block([Expr(e1); While(e2, Block([b; Expr(e3)]))])) *)
    | Ast.While (e, b) ->
        let b' = stmt env num_formals b and e' = expr env e in
        [Bra (1+ List.length b')] @ b' @ e' @
        [Bne (-(List.length b' + List.length e'))]
    | Ast.Foreach(e1, e2, b) -> 
        let id = (match e1 with Ast.Id id -> id | _ -> raise (Failure "foreach expects an indentifier")) in
	      let a' = (try [Sfp (StringMap.find id env.local_index)]
                 with Not_found -> try [Str (StringMap.find id env.global_index)]
                 with Not_found -> raise (Failure ("undeclared variable " ^ id))) in
        let b' = a' @ (stmt env num_formals b) @ [Drp] and e' = expr env e2 in
        e' @ [Bra (1+ List.length b')] @ b' @
        [Nit (-(List.length b'))]
	
        
  in
  (* Translate a function in AST form into a list of bytecode statements *)
  let translate env fdecl =
    (* Bookkeeping: FP offsets for locals and arguments *)
    let num_formals = List.length fdecl.Ast.formals
    and num_locals = List.length fdecl.Ast.locals
    and local_offsets = enum 1 1 fdecl.Ast.locals
    and formal_offsets = enum (-1) (-2) fdecl.Ast.formals in
    let env = { env with local_index = string_map_pairs
      StringMap.empty (local_offsets @ formal_offsets) }


    in [Ent num_locals] @      (* Entry: allocate space for locals *)
      stmt env num_formals (Ast.Block fdecl.Ast.body) @  (* Body *)
      [Lit (Types.Int 0); Rts num_formals]   (* Default = return 0 *)

  in let env = { function_index = function_indexes;
     global_index = global_indexes;
     local_index = StringMap.empty } in

  let translate_toplevel =
     (Ent 0) :: stmt env 0 (Ast.Block (List.rev (List.fold_left (fun s e-> (match e with 
         Ast.Stmt(st) -> st::s
       | _ -> s)) [] gstmts))) 
    @ [Lit (Types.Int 0); Rts 0]
  in

  (* Code executed to start the program: Jsr main; halt *)
  let entry_function = translate_toplevel(*try
    [Jsr (StringMap.find "main" function_indexes); Hlt] 
  with Not_found -> raise (Failure ("no \"main\" function"))*)
  in
    
  (* Compile the functions *)
  let func_bodies = entry_function :: List.map (translate env) functions in

  (* Calculate function entry points by adding their lengths *)
  let (fun_offset_list, _) = List.fold_left
      (fun (l,i) f -> (i :: l, (i + List.length f))) ([],2) func_bodies in
  let func_offset = Array.of_list (List.rev fun_offset_list) in

  { num_globals = List.length globals;
    (* Concatenate the compiled functions and replace the function
       indexes in Jsr statements with PC values *)
    text = Array.of_list ((Jsr 2) :: Hlt :: List.map (function
        Jsr i when i > 0 -> Jsr func_offset.(i)
      | Rsn i when i > 0 -> Rsn func_offset.(i)
      | _ as s -> s) (List.concat func_bodies))
  }
