open Ast
open Bytecode

exception SelectReturnException of bool

(* Stack layout just after "Ent":

              <-- SP
   Local n
   ...
   Local 0
   Saved FP   <-- FP
   Saved PC
   Arg 0
   ...
   Arg n *)

let boolean = function
    Types.Boolean(true) -> true
  | _                   -> false

let int_of_t = function
    Types.Int i -> i
  | _           -> raise (Failure "Invalid frame pointer")

let execute_prog prog =
  let stack = Array.make 1024 Types.Undef
  and globals = Array.make prog.num_globals Types.Undef in
  let rec listcons n top s = 
    if n == 0 then s
	  else listcons (n-1) (top-1) (stack.(top) :: s) in
	
  let rec hashcons n top s =
  	if n == 0 then s
    else hashcons (n-1) (top-2) ((Types.string_of_typed_expr stack.(top), stack.(top-1)) :: s) in
  let rec exec fp sp pc = match prog.text.(pc) with
    Lit i  -> stack.(sp) <- i ; exec fp (sp+1) (pc+1)
  | Drp    -> exec fp (sp-1) (pc+1)
  | Bin op -> let op1 = stack.(sp-2) and op2 = stack.(sp-1) in     
      stack.(sp-2) <- (let boolean = function 
          Types.Boolean(true) -> Types.Int 1
        | _                   -> Types.Int 0 in   
	    match op with
      	Add     -> Types.addition op1 op2
      | Sub     -> Types.subtraction op1 op2
      | Mult    -> Types.multiplication op1 op2
      | Div     -> Types.division op1 op2
      | Equal   -> boolean (Types.typed_eq op1 op2)
      | Neq     -> boolean (Types.typed_neq op1 op2)
      | Less    -> boolean (Types.typed_less op1 op2)
      | Leq     -> boolean (Types.typed_leq op1 op2)
      | Greater -> boolean (Types.typed_greater op1 op2)
      | Geq     -> boolean (Types.typed_geq op1 op2) );
      exec fp (sp-1) (pc+1)
  | Lod i   -> stack.(sp)   <- globals.(i)  ; exec fp (sp+1) (pc+1)
  | Str i   -> globals.(i)  <- stack.(sp-1) ; exec fp sp     (pc+1)
  | Lfp i   -> stack.(sp)   <- stack.(fp+i) ; exec fp (sp+1) (pc+1)
  | Sfp i   -> stack.(fp+i) <- stack.(sp-1) ; exec fp sp     (pc+1)
  | Jsr(-1) -> print_endline (Types.string_of_typed_expr stack.(sp-1)) ; exec fp sp (pc+1)
  | Jsr(-2) -> stack.(sp-1) <- Util.gets stack.(sp-1) ; exec fp sp (pc+1)
  | Jsr(-3) -> stack.(sp-1) <- Util.parse stack.(sp-1) ; exec fp sp (pc+1)
  | Jsr i   -> stack.(sp)   <- Types.Int (pc + 1)     ; exec fp (sp+1) i
  | Ent i   -> stack.(sp)   <- Types.Int fp           ; exec sp (sp+i+1) (pc+1)
  | Rts i   -> let new_fp = int_of_t stack.(fp) and new_pc = int_of_t stack.(fp-1) in
	       if new_pc < 0 then raise (SelectReturnException(Types.toBool(stack.(sp-1))))
               else stack.(fp-i-1) <- stack.(sp-1); exec new_fp (fp-i) new_pc
  | Beq i   -> exec fp (sp-1) (pc + if boolean (Types.typed_eq stack.(sp-1) (Types.Int 0)) then i else 1)
  | Bne i   -> exec fp (sp-1) (pc + if boolean (Types.typed_neq stack.(sp-1) (Types.Int 0)) then i else 1)
  | Bra i   -> exec fp sp (pc+i)
  | Lis i   -> stack.(sp-i) <- Types.Lis (listcons i (sp-1) []); 
               exec fp (sp-i+1) (pc+1)
  | Has i   -> stack.(sp-2*i) <- Types.Hash (hashcons i (sp-1) []); 
               exec fp (sp-2*i+1) (pc+1)
  | Nit i   -> (match stack.(sp-1) with
                     Types.Lis l -> (match l with
                                          []     -> exec fp (sp-1) (pc+1)
                                        | hd::tl -> stack.(sp) <- hd; stack.(sp-1) <- Types.Lis tl; exec fp (sp+1) (pc+i))
                   | _           -> raise (Failure "foreach expects a list"))
  | Rsn i   -> stack.(sp) <- Types.Int(i) ; exec fp (sp+1) (pc+1)
  | Sel s   -> stack.(sp) <- (match s with Ast.DoubleColon -> Types.Int(1) | _ -> Types.Int(0)); exec fp (sp+1) (pc+1)
  | Slc i   -> let obj = stack.(sp-1) in
	       let chaining = (hashcons i (sp-2) []) in 
	       let post = stack.(sp-2-2*i) in
	       let restrictions = (match stack.(sp-2-2*i-1) with Types.Lis(l) -> l | _ as t -> raise (Failure ("Invalid restriction list error" ^ (Types.string_of_type t)))) in
	       let rec select obj chain post restrict parent  =
		 (match chain with
		   | (chain_type,chain_typed)::chain_new ->
		     let chain_value = match chain_typed with Types.Str(o) -> o | _ as expr -> Types.string_of_typed_expr expr in
		     (match chain_type with
			 "0" -> (match obj with
			     Types.Lis(o) ->
			       (try select (List.nth o (int_of_string (Ast.removeQuotes chain_value))) chain_new post restrict obj with Failure exp -> [])
			   | Types.Hash(o) -> let (_,next) = (List.find (fun v -> chain_value = fst v) o) in
					      select next chain_new post restrict obj
			   | _ -> [])
		       | "1" -> (match obj with
			     Types.Lis(o) -> let res, _  = List.fold_left
					       (fun (acc, idx) v ->
						 let nacc =
						   select v (if (Ast.removeQuotes chain_value) = (string_of_int idx) then chain_new else chain) post restrict obj
						 in (nacc::acc, idx+1)) ([], 0) o
					     in List.concat (List.rev res)
			   | Types.Hash(o) -> let res = List.fold_left
						(fun acc (key, v) ->
						  let nacc =
						    select v (if chain_value = key then chain_new else chain) post restrict obj
						  in (nacc::acc)) [] o
					      in List.concat (List.rev res)
			   | _ -> [])
		       | _ -> raise (Failure ("invalid selector")))
		   | _ -> let check_restrictions = List.fold_left (fun isvalid f -> 
                              try
				stack.(sp) <- parent; stack.(sp+1) <- Types.Int(-1); exec fp (sp+2) (int_of_string (Types.string_of_typed_expr f)); true
			      with SelectReturnException(v) -> (v && isvalid)) true restrictions in
			  if check_restrictions then  (match post with
                              ">" -> [parent]
                            | _ -> [obj]) else [])
               in let out = select obj chaining (Types.string_of_typed_expr post) restrictions Types.Undef in
               let returnlist = (List.fold_left (fun isdouble (chain_type, _) -> (match chain_type with
                      "1" -> true
		    | _ -> isdouble || false)) false chaining) in
               let result = (match returnlist  with
                      false -> (match out with hd::tl -> hd | _ -> Types.Undef)
		    | true  -> Types.Lis(out)) in stack.(sp-(2*i+2+1)) <- result; exec fp (sp-(2*i+2+1)+1) (pc+1)
  | Hlt     -> ()

  in exec 0 0 0
