(*open Ast*)
open Jscanner

module NameMap = Map.Make(struct
    type t = string
    let compare x y = Pervasives.compare x y
end)

exception ReturnException of Types.t * Types.t NameMap.t

module StringMap = Map.Make(String)

let debug_map m = 
    let print_key k v = print_endline k in NameMap.iter print_key m

let rec exec toplevel funcs env = function
      Ast.Block(stmts) -> List.fold_left (exec toplevel funcs) env stmts
    | Ast.Expr(e) -> let _, env, funcs = eval env funcs toplevel e in env
    | Ast.If(e, s1, s2) ->
        let v, env, funcs = eval env funcs toplevel e in
        exec toplevel funcs env (if Types.toBool(v) then s1 else s2)
    | Ast.While(e, s) ->
        let rec loop env =
            let v, env, funcs = eval env funcs toplevel e in
                if Types.toBool(v) then loop (exec toplevel funcs env s) else env
        in loop env
    | Ast.Foreach(e1, e2, s) ->
      (match e1 with
        Ast.Id(id) ->
          let v, env, funcs = eval env funcs toplevel e2 in
          let rec loop env = function
              hd :: tl -> let locals, globals = env in 
                          let env = if toplevel then (locals, NameMap.add id hd globals) else (NameMap.add id hd locals, globals) in
                          let env = exec toplevel funcs env s in
                          loop env tl
            | _        -> env
          in (match v with
            Types.Lis l -> loop env l
	        | _ -> raise (Failure "foreach expects a list"))
      | _          -> raise (Failure "foreach expects an identifier"))

    | Ast.Return(e) -> if toplevel then raise (Failure "return cannot be used at top level") else
        let v, (locals, globals), funcs = eval env funcs toplevel e in
        raise (ReturnException(v, globals))

and

call fdecl actuals globals funcs =
    (* Execute a statement and return an updated environment *)
    let toplevel = false
    in
        (* Enter the function: bind actual values to formal arguments *)
        let locals =
            try List.fold_left2
                (fun locals formal actual -> NameMap.add formal actual locals)
                NameMap.empty fdecl.Ast.formals actuals
            with Invalid_argument(_) ->
                raise (Failure ("wrong number of arguments passed to " ^ fdecl.Ast.fname))
        in
            (* Execute each statement in sequence, return updated global symbol table *)
            snd (List.fold_left (exec toplevel funcs) (locals, globals) fdecl.Ast.body)

and

eval env funcs toplevel = function
      Ast.Literal(i) -> Types.Int(i), env, funcs
    | Ast.Float(f)   -> Types.Float(f), env, funcs
    | Ast.Str(s)     -> Types.Str(s), env, funcs (*Types.Str(removeQuotes s), env*)
    | Ast.Boolean(b) -> Types.Boolean(b), env, funcs
    | Ast.Undef      -> Types.Undef, env, funcs
    | Ast.Id(var)    ->
        let locals, globals = env in
        if NameMap.mem var locals then
            (NameMap.find var locals), env, funcs
        else if NameMap.mem var globals then
            (NameMap.find var globals), env, funcs
        else
            raise (Failure ("undeclared variable " ^ var))
    | Ast.Binop(e1, op, e2) ->
        let v1, env, funcs = eval env funcs toplevel e1 in
        let v2, env, funcs = eval env funcs toplevel e2 in
    (match op with
        Ast.Add     -> Types.addition v1 v2
      | Ast.Sub     -> Types.subtraction v1 v2
      | Ast.Mult    -> Types.multiplication v1 v2
      | Ast.Div     -> Types.division v1 v2
      | Ast.Equal   -> Types.typed_eq v1 v2
      | Ast.Neq     -> Types.typed_neq v1 v2
      | Ast.Less    -> Types.typed_less v1 v2
      | Ast.Leq     -> Types.typed_leq v1 v2
      | Ast.Greater -> Types.typed_greater v1 v2
      | Ast.Geq     -> Types.typed_geq v1 v2), env, funcs
      | Ast.Assign(var, expr) -> let v, (locals, globals), funcs = eval env funcs toplevel expr in
        if NameMap.mem var locals then
            v, (NameMap.add var v locals, globals), funcs
        else if NameMap.mem var globals then
            v, (locals, NameMap.add var v globals), funcs
        else if toplevel then
            v, (locals, NameMap.add var v globals), funcs       
  else
            v, (NameMap.add var v locals, globals), funcs
            (*raise (Failure ("undeclared identifier " ^ var))*)
    | Ast.Lis(list) -> let (res, env, funcs) = 
       (List.fold_left (fun (acc, env, funcs) x -> 
         let out, new_env, new_funcs = eval env funcs toplevel x in 
         (out::acc, new_env, new_funcs)) ([], env, funcs) list) in 
           Types.Lis(List.rev res), env, funcs
    | Ast.Hash(hash) -> let (res, env, funcs) = 
        (List.fold_left (fun (map, env, funcs) (x,y) -> 
          let outx = x (*removeQuotes x*) in
          let outy, env, funcs = eval env funcs toplevel y in 
          ((outx,outy)::(List.filter ((fun new_key (m,n) -> new_key <> m) outx) map), env, funcs)) ([], env, funcs) hash) 
      in Types.Hash(List.rev res), env, funcs 
    | Ast.Select(id, chaining, postfix, restrictions) -> 
      let rec select obj chain post restrict parent env funcs =
	match chain with
	  | (chain_type,chain_expr)::chain_new ->
	    let chain_typed,_,_ = eval env funcs toplevel chain_expr in
	    let chain_value = match chain_typed with Types.Str(o) -> o | _ as expr -> Types.string_of_typed_expr expr in
	    (match chain_type with
		Ast.Colon -> (match obj with
		    Types.Lis(o) -> 
		      (try select (List.nth o (int_of_string (Ast.removeQuotes chain_value))) chain_new post restrict obj env funcs with Failure exp -> [], env, funcs)
		  | Types.Hash(o) -> let (_,next) = (List.find (fun v -> chain_value = fst v) o) in
		    select next chain_new post restrict obj env funcs
		  | _ -> [], env, funcs)
	      | Ast.DoubleColon -> (match obj with
		    Types.Lis(o) -> let res, env, funcs, _ = List.fold_left 
				    (fun (acc, env, funcs, idx) v -> 
				      let  nacc, nenv, nfuncs = 
					select v (if (Ast.removeQuotes chain_value) = (string_of_int idx) then chain_new else chain) post restrict obj env funcs 
				      in (nacc::acc, env, funcs, idx+1)) ([], env, funcs, 0) o 
				  in List.concat (List.rev res), env, funcs
		  | Types.Hash(o) -> let res, env, funcs = List.fold_left 
				    (fun (acc, env, funcs) (key, v) -> 
				      let nacc, nenv, nfuncs = 
					select v (if chain_value = key then chain_new else chain) post restrict obj env funcs 
				      in (nacc::acc, env, funcs)) ([], env, funcs) o 
				  in List.concat (List.rev res), env, funcs
		  | _ -> [], env, funcs))
	      | _ -> let check_restrictions, env, funcs = List.fold_left (fun (isvalid, env, funcs) f -> 
		let fdecl =
		  try NameMap.find f funcs
		  with Not_found -> raise (Failure ("undefined function " ^ f))
		in let (locals, globals) = env in
                   try
                     let globals = call fdecl [parent] globals funcs
                     in (isvalid, (locals, globals), funcs)
                   with ReturnException(v, globals) -> ((Types.toBool(v) && isvalid), (locals,globals), funcs)) (true,env,funcs) restrictions in 
		      if check_restrictions then  (match post with
			  ">" -> [parent], env, funcs
			| _ -> [obj], env, funcs) else [], env, funcs
	     in let expr, env, funcs = eval env funcs toplevel id in
		let out, env, funcs = select expr chaining postfix restrictions Types.Undef env funcs in 
		let returnlist = (List.fold_left (fun isdouble (chain_type, _) -> 
		  (match chain_type with
		      Ast.DoubleColon -> true
		    | _ -> isdouble || false)) false chaining) in
		(match returnlist  with
		    false -> (match out with hd::tl -> hd | _ -> Types.Undef), env, funcs
	      | true  -> Types.Lis(out), env, funcs)
    | Ast.Call("print", [expr]) -> let res, env, funcs = eval env funcs toplevel expr in print_endline (Types.string_of_typed_expr res); 
        Types.Int(0), env, funcs
    | Ast.Call("gets", [expr]) -> let res, env, funcs = eval env funcs toplevel expr in 
	                                let res = Util.gets res in
                                  res, env, funcs
    | Ast.Call("parse", [expr])-> let res, env, funcs = eval env funcs toplevel expr in
                                  let res = Util.parse res in res, env, funcs
    | Ast.Call(f, actuals) ->
        (*print_endline ("call function " ^ f);*)
        let fdecl =
            try NameMap.find f funcs
            with Not_found -> raise (Failure ("undefined function " ^ f))
        in (*let fdecl*)
	let actuals, env = List.fold_left
                (fun (actuals, env) actual ->
                    let v, env, funcs = eval env funcs toplevel actual in v :: actuals, env)
                ([], env) (List.rev actuals)
            in (*let actuals, env*)
                (*List.map (fun a -> print_endline (Types.string_of_typed_expr a)) actuals;*)
                let (locals, globals) = env in
                try
                    let globals = call fdecl actuals globals funcs
                    in Types.Int(0), (locals, globals), funcs
                with ReturnException(v, globals) -> v, (locals, globals), funcs

let rec run_gstmt env funcs = function
      Ast.Stmt(s) -> let env = exec true funcs env s in env, funcs
(* (match s with
           Ast.Expr(e)      -> let res, env, funcs = eval env funcs true e in env, funcs
         | Ast.Block(stmts) -> List.fold_left (fun (env, funcs) e -> run_gstmt env funcs (Ast.Stmt e)) (env, funcs) stmts
         
         | _                -> raise (Failure "unsupported statement")
          (*Ast.If...*)
          (*Ast.While...*)
    )  *)
    | Ast.Func(fdecl) ->
        let funcs = NameMap.add fdecl.Ast.fname fdecl funcs in
        env, funcs

let run program = 
    let env = (NameMap.empty, NameMap.empty) in
    let funcs = NameMap.empty in
    List.fold_left (fun (e, f) stmt -> run_gstmt e f stmt) (env, funcs) program

(*
let _ =
    let lexbuf = Lexing.from_channel stdin in
    let program = Parser.program Scanner.token lexbuf in
    let env = (NameMap.empty, NameMap.empty) in
    let funcs = NameMap.empty in
    List.fold_left (fun (e, f) stmt -> run e f stmt) (env, funcs) (List.rev program)
*)
