
exception UninitializedVariable of string
exception FunctionNotDefined of string
exception MultipleDefinition of string
exception UnboundVariable of string

type expression = Int | Float | Str | Boolean | Lis | Hash | Any | Unbound

type symbol_table = {
  parent : symbol_table option;
  variables : (string * expression) list
}


(***from compile.ml**)
module NameMap = Map.Make(struct
    type t = string
    let compare x y = Pervasives.compare x y
end)

module StringMap = Map.Make(String)

(*
  implement the scope rule
  an assignment at toplevel creates a global variable if it is not defined yet.
  an assignment at a fun_decl, however, creates a local variable when the left
  hand side is neither a global nor formal let variable.
  a formal variable shadows a global variable of the same name
*)
let rec env_expr env formals toplevel = function
      Ast.Assign(var, expr) -> let globals, locals, funcs = env in
                               if List.mem var formals || List.mem var globals then
                                   env
                               else if toplevel then
                                   var :: globals, locals, funcs
                               else
                                   globals, var :: locals, funcs
    | _                     -> env

and env_stmt env formals toplevel = function
      Ast.Expr(e)       -> let env = env_expr env formals toplevel e in env
    | Ast.Block(stmts)  -> List.fold_left (fun env s -> env_stmt env formals toplevel s) env stmts
    | Ast.If(e,s1,s2)   -> let env = env_expr env formals toplevel e in 
                           let env = env_stmt env formals toplevel s1 in
                           let env = env_stmt env formals toplevel s2 in
                           env
    | Ast.While(e,s)    -> let env = env_expr env formals toplevel e in
                           let env = env_stmt env formals toplevel s in env
    | Ast.Foreach(e1,e2,s) -> let globals, locals, funcs = env in
                           let env = (match e1 with 
                               Ast.Id id -> if toplevel then id::globals, locals, funcs
	                                          else globals, id :: locals, funcs
											       | _ -> raise (Failure "foreach expects an identifier"))
  		                     in env_stmt env formals toplevel s
    | _                 -> env

and env_gstmt env = function
      Ast.Stmt(s)     -> let env = env_stmt env [] true s in env
    | Ast.Func(fdecl) -> let env = env_fdecl env fdecl.Ast.formals fdecl in env

and env_fdecl env formals fdecl = 
    let _,_,funcs = env in
      if NameMap.mem fdecl.Ast.fname funcs then raise (MultipleDefinition fdecl.Ast.fname)
      else
        let globals, localvars, funcs = List.fold_left (fun env s -> env_stmt env formals false s) env fdecl.Ast.body in 
        let funcs = NameMap.add fdecl.Ast.fname {fdecl with Ast.locals=localvars} funcs in
        globals, [], funcs
(**end compile.ml**)

let fun_exists f funcs = 
	try
		ignore(List.find (fun e -> e.Ast.fname = f) funcs); true
	with Not_found -> raise (FunctionNotDefined f)

let rec find_variable (scope : symbol_table) name = 
	  try
	    List.find (fun (s, _) -> s = name) scope.variables
	  with Not_found -> 
	    match scope.parent with
	      Some(parent) -> find_variable parent name
	    | _ -> raise (UnboundVariable name)

(* TODO add a variable to symble table *)
let rec add_variable env (name, t) = 
  let sym = fst env in
  let funcs = snd env in
  let sym = {parent = sym.parent; variables = (name, t) :: sym.variables} in
  (sym, funcs) 

let rec expr ((sym, funcs) as env) toplevel = function
    Ast.Literal(i) -> env, Int
  | Ast.Float(f)   -> env, Float
  | Ast.Str(s)     -> env, Str
  | Ast.Boolean(b) -> env, Boolean
  | Ast.Lis(l)     -> env, Lis
  | Ast.Hash(h)    -> env, Hash
  | Ast.Id(var)    -> let _, t = find_variable sym var in env, t(*TODO perform lookup in env, raise an exception if var is uninitialized*)
  | Ast.Binop(e1, op, e2) -> let env, _ = expr env toplevel e1 in let env, _ = expr env toplevel e2 in env, Any (*TODO*)
  | Ast.Assign(var, e) -> let env, t = expr env toplevel e in let env = add_variable env (var, t) in env, Any (*TODO add var to env with the type of expr*)
	| Ast.Select(var,_,_,_) -> env, Any (*TODO var has to be a list, a hash, or an "any"*)
  | Ast.Call(f, actuals)  -> ignore(fun_exists f funcs); env, Any (*TODO return type of f?*)
  | Ast.Undef             -> env, Unbound (* TODO Shall we raise an exception *)

(*TODO the "intersection" of two enviroment*)
let merge env1 env2 =
  let sym1 = fst env1 in
  let sym2 = fst env2 in
  let v = List.filter (fun x -> List.mem x sym2.variables) sym1.variables in
  ({parent = sym1.parent; variables = v}, snd env1)
	
let rec stmt env toplevel = function
    Ast.Block(stmts)  -> List.fold_left (fun e s -> stmt env toplevel s) env stmts 
  | Ast.Expr e        -> let env, _ = expr env toplevel e in env
  | Ast.If(e, s1, s2) -> (*how to handle if statement*)
	                      let env, _ = expr env toplevel e in 
                        let env1 = stmt env toplevel s1 in
                        let env2 = stmt env toplevel s2 in
                        merge env1 env2
  | Ast.While(e, s)  -> let env1,_ = expr env toplevel e in
                        let env2 = stmt env1 toplevel s in
                        merge env1 env2
  | Ast.Return(e)    -> (*TODO if all return statements agree in type, put the return type to env*)
                        env
  | Ast.Foreach(e1, e2, s) -> let id = (match e1 with Ast.Id id -> id | _ -> raise (Failure "foreach expects an identifier")) in
                              let env, t = expr env toplevel e2 in
                              let env1 = add_variable env (id, Any) in
                              let env2 = stmt env1 toplevel s in
                              merge env1 env2;;

let func env toplevel f = 
  let func = List.find (fun x -> x.Ast.fname = f.Ast.fname) (snd env) in
  let vars = List.fold_left (fun l v -> (v, Any) :: l) [] func.Ast.formals in
  let vars = List.fold_left (fun l v -> (v, Any) :: l) vars func.Ast.locals in
	let sym = {parent = Some((fst env)); variables = vars} in
  stmt (sym, (snd env)) toplevel (Ast.Block f.Ast.body)

let rec gstmt env = function
    Ast.Stmt s -> stmt env true s
  | Ast.Func f -> func env false f;;

let check program =
  let functions = NameMap.add "print" {Ast.fname="print"; Ast.formals=[]; Ast.locals=[]; Ast.body=[]} NameMap.empty in
  let functions = NameMap.add "gets" {Ast.fname="gets"; Ast.formals=[]; Ast.locals=[]; Ast.body=[]} functions in
  let functions = NameMap.add "parse" {Ast.fname="parse"; Ast.formals=[]; Ast.locals=[]; Ast.body=[]} functions in	
	let (globals, _, functions) = List.fold_left env_gstmt ([], [], functions) program in
  let funcs = NameMap.fold (fun k v a -> v :: a) functions [] in
  let sym = {parent = None; variables = List.fold_left (fun l v -> (v, Any) :: l) [] globals} in 
	(*let funcs = List.fold_left (fun s e -> match e with
		  Ast.Stmt st -> s
    | Ast.Func f -> if List.mem f.Ast.fname s then raise (MultipleDefinition f.Ast.fname) else f.Ast.fname :: s) [] program in*)
	List.fold_left (fun env gs -> gstmt env gs) (sym, funcs) program
