type t = 
   Undef (* for compiler. when all local variables are calculated in advance, a variable is undefined before the decl*)
 | Int of int
 | Float of float
 | Str of string
 | Boolean of bool
 | Lis of t list 
 | Hash of (string*t) list

let string_of_type = function
      Int(v) -> "Integer"
    | Float(v) -> "Float"
    | Str(v) -> "String"
    | Boolean(v) -> "Boolean"
    | Lis(v) -> "List"
    | Hash(v) -> "Map"
    | Undef -> "Undefined"

let addition a b = 
    match a, b with
      Int(i1), Int(i2)     -> Int (i1 + i2)
    | Float(f1), Float(f2) -> Float (f1 +. f2)
    | Int(i1), Float(f2)   -> Float ((float_of_int i1) +. f2)
    | Float(f1), Int(i2)   -> Float (f1 +. (float_of_int i2))
    | Str(s1), Str(s2)     -> Str(s1 ^ s2)
    | Str(s1), Int(i2)     -> Str(s1^ (string_of_int i2))
    | Int(i1), Str(s2)     -> Str((string_of_int i1) ^ s2)
    | Str(s1), Float(f2)   -> Str(s1 ^ (string_of_float f2))
    | Float(f1), Str(s2)   -> Str((string_of_float f1) ^ s2)
    | Hash(f1), Hash(f2)   -> Hash(f1 @ f2) (*more work here to preserve the uniqness of the keys*)
    | Lis(l1), Lis(l2)     -> Lis(l1 @ l2)
    | _ -> raise(Failure("can't add " ^ string_of_type a ^ " and "  ^ string_of_type b))
        

let subtraction a b = 
    match a, b with
      Int(i1), Int(i2)     -> Int (i1 - i2)
    | Float(f1), Float(f2) -> Float (f1 -. f2)
    | Int(i1), Float(f2)   -> Float ((float_of_int i1) -. f2)
    | Float(f1), Int(i2)   -> Float (f1 -. (float_of_int i2))
    | Lis(l1), Lis(l2) -> Lis(List.filter (fun x -> not (List.mem x l2)) l1)
    | _ -> raise(Failure("can't subtract " ^ string_of_type a ^ " and "  ^ string_of_type b))

let multiplication a b = 
    match a, b with
      Int(i1), Int(i2)     -> Int (i1 * i2)
    | Float(f1), Float(f2) -> Float (f1 *. f2)
    | Int(i1), Float(f2)   -> Float ((float_of_int i1) *. f2)
    | Float(f1), Int(i2)   -> Float (f1 *. (float_of_int i2))
    | _ -> raise(Failure("can't multiply " ^ string_of_type a ^ " and "  ^ string_of_type b))

let division a b = 
    match a, b with
      Int(i1), Int(i2)     -> Int (i1 / i2)
    | Float(f1), Float(f2) -> Float (f1 /. f2)
    | Int(i1), Float(f2)   -> Float ((float_of_int i1) /. f2)
    | Float(f1), Int(i2)   -> Float (f1 /. (float_of_int i2))
    | _ -> raise(Failure("can't divide " ^ string_of_type a ^ " and "  ^ string_of_type b))


let toBool a = match a with
      Int(v)     -> v != 0
    | Float(v)   -> v != 0.0
    | Str(v)     -> v != ""
    | Boolean(v) -> v
    | Undef -> false
    | _ -> raise(Failure(string_of_type a ^ " can't is an invalid boolean expression"))

let typed_eq v1 v2 = match v1, v2 with 
      Int(i1), Int(i2) -> Boolean(i1 = i2)
    | Float(f1), Float(f2) -> Boolean(f1 = f2)
    | Str(s1), Str(s2) -> Boolean(s1 = s2)
    | Boolean(b1), Boolean(b2) -> Boolean(b1 = b2)
    | _ -> Boolean(false)

let typed_neq v1 v2 = match v1, v2 with 
      Int(i1), Int(i2) -> Boolean(i1 <> i2)
    | Float(f1), Float(f2) -> Boolean(f1 <> f2)
    | Str(s1), Str(s2) -> Boolean(s1 <> s2)
    | Boolean(b1), Boolean(b2) -> Boolean(b1 <> b2)
    | _ -> Boolean(true)

let typed_less v1 v2 =   match v1, v2 with
      Int(i1), Int(i2) -> Boolean(i1 < i2)
    | Float(f1), Float(f2) -> Boolean(f1 < f2)
    | _ -> Boolean(false)

let typed_leq v1 v2 = match  v1, v2 with
      Int(i1), Int(i2) -> Boolean(i1 <= i2)
    | Float(f1), Float(f2) -> Boolean(f1 <= f2)
    | _ -> Boolean(false)

let typed_greater v1 v2 = match v1, v2 with
      Int(i1), Int(i2) -> Boolean(i1 > i2)
    | Float(f1), Float(f2) -> Boolean(f1 > f2)
    | _ -> Boolean(false)

let typed_geq v1 v2 = match v1, v2 with
      Int(i1), Int(i2) -> Boolean(i1 >= i2)
    | Float(f1), Float(f2) -> Boolean(f1 >= f2)
    | _ -> Boolean(false)


let rec string_of_typed_expr = function
      Int(v)     -> string_of_int v
    | Float(v)   -> string_of_float v
    | Str(v)     -> v (*"\""^v^"\""*)
    | Boolean(v) -> string_of_bool v
    | Lis(v)     -> "[" ^ String.concat "," (List.map string_of_typed_expr v) ^ "]"
    | Hash(v)    -> let string_of_kvpair = fun (k, v) -> k ^ ":" ^ string_of_typed_expr v in "{" ^ String.concat "," (List.map string_of_kvpair v)  ^ "}"
    | Undef      -> "null"

